from google.protobuf import empty_pb2

from django_grpc_framework.protobuf.json_format import (
    message_to_dict, parse_dict
)


class CreateModelMixin:
    """
    Create a model instance.
    """
    def Create(self, request, context):
        data = message_to_dict(request)
        serializer = self.get_serializer(data=data)
        serializer.is_valid(raise_exception=True)
        self.perform_create(serializer)
        protobuf_class = self.get_protobuf_class()
        return parse_dict(serializer.data, protobuf_class())

    def perform_create(self, serializer):
        serializer.save()


class ListModelMixin:
    """
    List a queryset.
    """
    def List(self, request, context):
        queryset = self.filter_queryset(self.get_queryset())
        serializer = self.get_serializer(queryset, many=True)
        protobuf_class = self.get_protobuf_class()
        for data in serializer.data:
            yield parse_dict(data, protobuf_class())


class RetrieveModelMixin:
    """
    Retrieve a model instance.
    """
    def Retrieve(self, request, context):
        instance = self.get_object()
        serializer = self.get_serializer(instance)
        protobuf_class = self.get_protobuf_class()
        return parse_dict(serializer.data, protobuf_class())


class UpdateModelMixin:
    """
    Update a model instance.
    """
    def Update(self, request, context):
        instance = self.get_object()
        data = message_to_dict(request)
        serializer = self.get_serializer(instance, data=data)
        serializer.is_valid(raise_exception=True)
        self.perform_update(serializer)

        if getattr(instance, '_prefetched_objects_cache', None):
            # If 'prefetch_related' has been applied to a queryset, we need to
            # forcibly invalidate the prefetch cache on the instance.
            instance._prefetched_objects_cache = {}

        protobuf_class = self.get_protobuf_class()
        return parse_dict(serializer.data, protobuf_class())

    def perform_update(self, serializer):
        serializer.save()


class DestroyModelMixin:
    """
    Destroy a model instance.
    """
    def Destroy(self, request, context):
        instance = self.get_object()
        self.perform_destroy(instance)
        return empty_pb2.Empty()

    def perform_destroy(self, instance):
        instance.delete()



from django.db.models.query import QuerySet
from django.shortcuts import get_object_or_404
from django.core.exceptions import ValidationError
from django.http import Http404
import grpc

from django_grpc_framework import mixins, services


class GenericService(services.Service):
    """
    Base class for all other generic services.
    """
    # Either set this attribute or override ``get_queryset()``.
    queryset = None
    # Either set this attribute or override ``get_serializer_class()``.
    serializer_class = None
    # Either set this attribute or override ``get_protobuf_class``.
    protobuf_class = None
    # Set this if you want to use object lookups other than id
    lookup_field = 'id'
    lookup_request_field = None

    def get_queryset(self):
        """
        Get the list of items for this service.
        This must be an iterable, and may be a queryset.
        Defaults to using ``self.queryset``.

        If you are overriding a handler method, it is important that you call
        ``get_queryset()`` instead of accessing the ``queryset`` attribute as
        ``queryset`` will get evaluated only once.
        """
        assert self.queryset is not None, (
            "'%s' should either include a ``queryset`` attribute, "
            "or override the ``get_queryset()`` method."
            % self.__class__.__name__
        )
        queryset = self.queryset
        if isinstance(queryset, QuerySet):
            # Ensure queryset is re-evaluated on each request.
            queryset = queryset.all()
        return queryset

    def get_serializer_class(self):
        """
        Return the class to use for the serializer.
        Defaults to using `self.serializer_class`.
        """
        assert self.serializer_class is not None, (
            "'%s' should either include a `serializer_class` attribute, "
            "or override the `get_serializer_class()` method."
            % self.__class__.__name__
        )
        return self.serializer_class

    def get_protobuf_class(self):
        """
        Return the class to use for the protobuf message.
        Defaults to using `self.protobuf_class`.
        """
        assert self.protobuf_class is not None, (
            "'%s' should either include a `protobuf_class` attribute, "
            "or override the `get_protobuf_class()` method."
            % self.__class__.__name__
        )
        return self.protobuf_class

    def get_object(self):
        queryset = self.filter_queryset(self.get_queryset())
        lookup_request_field = self.lookup_request_field or self.lookup_field
        assert hasattr(self.request, lookup_request_field), (
            'Expected service %s to be called with request that has a field '
            'named "%s". Fix your request protocol definition, or set the '
            '`.lookup_field` attribute on the service correctly.' %
            (self.__class__.__name__, lookup_request_field)
        )
        lookup_value = getattr(self.request, lookup_request_field)
        filter_kwargs = {self.lookup_field: lookup_value}
        try:
            return get_object_or_404(queryset, **filter_kwargs)
        except (TypeError, ValueError, ValidationError, Http404):
            self.context.abort(grpc.StatusCode.NOT_FOUND, (
                '%s: %s not found!' %
                (queryset.model.__name__, lookup_value)
            ))

    def get_serializer(self, *args, **kwargs):
        """
        Return the serializer instance that should be used for validating and
        deserializing input, and for serializing output.
        """
        serializer_class = self.get_serializer_class()
        kwargs.setdefault('context', self.get_serializer_context())
        return serializer_class(*args, **kwargs)

    def get_serializer_context(self):
        """
        Extra context provided to the serializer class.
        """
        return {
            'grpc_request': self.request,
            'grpc_context': self.context,
        }

    def filter_queryset(self, queryset):
        return queryset


class CreateService(mixins.CreateModelMixin,
                    GenericService):
    """
    Concrete service for creating a model instance.
    """
    pass


class ListService(mixins.ListModelMixin,
                  GenericService):
    """
    Concrete service for listing a queryset.
    """
    pass


class RetrieveService(mixins.RetrieveModelMixin,
                      GenericService):
    """
    Concrete service for retrieving a model instance.
    """
    pass


class DestroyService(mixins.DestroyModelMixin,
                     GenericService):
    """
    Concrete service for deleting a model instance.
    """
    pass


class UpdateService(mixins.UpdateModelMixin,
                    GenericService):
    """
    Concrete service for updating a model instance.
    """
    pass


class ReadOnlyModelService(mixins.RetrieveModelMixin,
                           mixins.ListModelMixin,
                           GenericService):
    """
    Concrete service that provides default ``List()`` and ``Retrieve()``
    handlers.
    """
    pass


class ModelService(mixins.CreateModelMixin,
                   mixins.RetrieveModelMixin,
                   mixins.UpdateModelMixin,
                   mixins.DestroyModelMixin,
                   mixins.ListModelMixin,
                   GenericService):
    """
    Concrete service that provides default ``Create()``, ``Retrieve()``,
    ``Update()``, ``Destroy()`` and ``List()`` handlers.
    """
    pass



from functools import update_wrapper

import grpc
from django.db.models.query import QuerySet
from django import db


class Service:
    def __init__(self, **kwargs):
        for key, value in kwargs.items():
            setattr(self, key, value)

    @classmethod
    def as_servicer(cls, **initkwargs):
        for key in initkwargs:
            if not hasattr(cls, key):
                raise TypeError(
                    "%s() received an invalid keyword %r. as_servicer only "
                    "accepts arguments that are already attributes of the "
                    "class." % (cls.__name__, key)
                )
        if isinstance(getattr(cls, 'queryset', None), QuerySet):
            def force_evaluation():
                raise RuntimeError(
                    'Do not evaluate the `.queryset` attribute directly, '
                    'as the result will be cached and reused between requests.'
                    ' Use `.all()` or call `.get_queryset()` instead.'
                )
            cls.queryset._fetch_all = force_evaluation

        class Servicer:
            def __getattr__(self, action):
                if not hasattr(cls, action):
                    return not_implemented

                def handler(request, context):
                    # db connection state managed similarly to the wsgi handler
                    db.reset_queries()
                    db.close_old_connections()
                    try:
                        self = cls(**initkwargs)
                        self.request = request
                        self.context = context
                        self.action = action
                        return getattr(self, action)(request, context)
                    finally:
                        db.close_old_connections()
                update_wrapper(handler, getattr(cls, action))
                return handler
        update_wrapper(Servicer, cls, updated=())
        return Servicer()


def not_implemented(request, context):
    """Method not implemented"""
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')



from google.protobuf.json_format import MessageToDict, ParseDict


def message_to_dict(message, **kwargs):
    kwargs.setdefault('including_default_value_fields', True)
    kwargs.setdefault('preserving_proto_field_name', True)
    return MessageToDict(message, **kwargs)


def parse_dict(js_dict, message, **kwargs):
    kwargs.setdefault('ignore_unknown_fields', True)
    return ParseDict(js_dict, message, **kwargs)



# -*- coding: utf-8 -*-
from concurrent import futures
from datetime import datetime
import sys
import errno
import os

import grpc
from django.utils import autoreload
from django.conf import settings
from django.core.management.base import BaseCommand, CommandError

from django_grpc_framework.settings import grpc_settings


class Command(BaseCommand):
    help = 'Starts a gRPC server.'

    # Validation is called explicitly each time the server is reloaded.
    requires_system_checks = False

    def add_arguments(self, parser):
        parser.add_argument(
            'address', nargs='?', default='[::]:50051',
            help='Optional address for which to open a port.'
        )
        parser.add_argument(
            '--max-workers', type=int, default=10, dest='max_workers',
            help='Number of maximum worker threads.'
        )
        parser.add_argument(
            '--dev', action='store_true', dest='development_mode',
            help=(
                'Run the server in development mode.  This tells Django to use '
                'the auto-reloader and run checks.'
            )
        )

    def handle(self, *args, **options):
        self.address = options['address']
        self.development_mode = options['development_mode']
        self.max_workers = options['max_workers']
        self.run(**options)

    def run(self, **options):
        """Run the server, using the autoreloader if needed."""
        if self.development_mode:
            if hasattr(autoreload, "run_with_reloader"):
                autoreload.run_with_reloader(self.inner_run, **options)
            else:
                autoreload.main(self.inner_run, None, options)
        else:
            self.stdout.write((
                "Starting gRPC server at %(address)s\n"
            ) % {
                "address": self.address,
            })
            self._serve()

    def _serve(self):
        server = grpc.server(futures.ThreadPoolExecutor(max_workers=self.max_workers),
                             interceptors=grpc_settings.SERVER_INTERCEPTORS)
        grpc_settings.ROOT_HANDLERS_HOOK(server)
        server.add_insecure_port(self.address)
        server.start()
        server.wait_for_termination()

    def inner_run(self, *args, **options):
        # If an exception was silenced in ManagementUtility.execute in order
        # to be raised in the child process, raise it now.
        autoreload.raise_last_exception()

        self.stdout.write("Performing system checks...\n\n")
        self.check(display_num_errors=True)
        # Need to check migrations here, so can't use the
        # requires_migrations_check attribute.
        self.check_migrations()
        now = datetime.now().strftime('%B %d, %Y - %X')
        self.stdout.write(now)
        quit_command = 'CTRL-BREAK' if sys.platform == 'win32' else 'CONTROL-C'
        self.stdout.write((
            "Django version %(version)s, using settings %(settings)r\n"
            "Starting development server at %(address)s\n"
            "Quit the server with %(quit_command)s.\n"
        ) % {
            "version": self.get_version(),
            "settings": settings.SETTINGS_MODULE,
            "address": self.address,
            "quit_command": quit_command,
        })
        try:
            self._serve()
        except OSError as e:
            # Use helpful error messages instead of ugly tracebacks.
            ERRORS = {
                errno.EACCES: "You don't have permission to access that port.",
                errno.EADDRINUSE: "That port is already in use.",
                errno.EADDRNOTAVAIL: "That IP address can't be assigned to.",
            }
            try:
                error_text = ERRORS[e.errno]
            except KeyError:
                error_text = e
            self.stderr.write("Error: %s" % error_text)
            # Need to use an OS exit because sys.exit doesn't work in a thread
            os._exit(1)
        except KeyboardInterrupt:
            sys.exit(0)
